/* vm.h */

void vm_init(void);
void vm_load(char *);
int vm_spawn(char *);
void vm_run(void);
void vm_register_signal_fd(int /* fd */, int /* queue */);

#define VM_MAXQUEUES    512
#define VM_RUNQ         0
#define VM_TIMEQ        1

#define VM_MOUSEQ	6
#define VM_BEATQ	7
#define VM_MIDIQ	8
#define VM_SOCKQ	9

#define VM_USERQMIN     10
#define VM_NOQUEUE      (-1)

